/*
 * SkrTokenizer.java
 *
 * Created on 2007/10/09, 9:44
 */

package com.kujirahand.sakura;

import com.kujirahand.sakura.command.*;
import com.kujirahand.utils.*;
import java.util.Vector;
import java.util.Hashtable;
/**
 *
 * @author kujira
 */
public class SkrTokenizer extends Tokenizer {
    
    public String filename;
    public int lineno;
    public int fileno;
    public SCommands commands;
    private Hashtable table;
    
    public SkrTokenizer(String source, String filename) {
        super(source);
        initCommand();
        filename = filename;
        lineno = 0;
        convertSutoton();
        commands = new SCommands();
    }
    
    private void initCommand() {
        table = new Hashtable();
        table.put("TR", SCommandTrack.class);
        table.put("Track", SCommandTrack.class);
        table.put("Tempo",SCommandTempo.class);
    }
    
    public void convertSutoton() {
        String res = "";
        SutotonList sutoton = new SutotonList();
        // convert
        this.position = 0;
        while (!EOS()) {
            char c = getCur();
            if (c < 256) {
                res += c;
                position++;
                continue;
            }
            // check sutoton
            boolean b_replace = false;
            for (int i = 0; i < sutoton.size(); i++) {
                SutotonItem item = (SutotonItem)sutoton.get(i);
                if (startsWith(item.sutoton)) {
                    res += item.mml;
                    position += item.len;
                    b_replace = true;
                    break;
                }
            }
            if (!b_replace) {
                res += c;
                position++;
                continue;
            }
        }
        // reset
        System.out.println(res);
        this.setSource(res);
    }
    
    // [a-z] || [A-Za-z_][A-Za-z_0-9]+ || 全角
    public String getCommand() {
        String res = "";
        // first letter
        if (!EOS()) {
            char c = getChar();
            if (isLowerCase(c)) {
                res += c;
                return res;
            }
            if (isUpperCase(c) || c == '_' || c >= 256) {
                res += c;
            } else {
                back(1);
                return "";
            }
        }
        // second letter
        while (!EOS()) {
            char c = getChar();
            if (isAlphaNum(c) || c == '_') {
                res += c;
                continue;
            }
            back(1);
            break;
        }
        return res;
    }
    
    public SCommands split()
        throws ParseException {
        while (!EOS()) {
            skipSpace();
            char c = getCur();
            if (c > 256) {
                throw new ParseException(Lang.PARSER_FOUND_UNKNOWN_WORD, "" + c, fileno, lineno);
            }
            // command
            if (isUpperCase(c)) {
                String cmd = getCommand();
                Class klass = (Class)table.get(cmd);
                if (klass == null) {
                    throw new ParseException(Lang.PARSER_FOUND_UNKNOWN_WORD, cmd, fileno, lineno);
                }
                SCommand scmd;
                try {
                    scmd = (SCommand)klass.newInstance();
                } catch (Exception e) {
                    throw new ParseException(Lang.SYSTEM_ERROR, cmd, fileno, lineno);
                }
                scmd.getParams(this, cmd);
                commands.add(scmd);
                continue;
            }
            switch (c) {
                case '\r': 
                    next(1);
                    break;
                case '\n':
                    lineno++;
                    next(1);
                    break;
                // note data
                case 'a':
                case 'b':
                case 'c':
                case 'd':
                case 'e':
                case 'f':
                case 'g':
                    parseNote(c);
                    break;
                // rest
                case 'r':
                    parseRest();
                    break;
                case 'o':
                    next(1);
                    parseOctave("o");
                    break;
                case 'v':
                    next(1);
                    parseVelocity("v");
                    break;
                case 'l':
                    next(1);
                    parseLength("l");
                    break;
                case '@':
                    next(1);
                    parseProgramChange("@");
                    break;
                case '>':
                    next(1);
                    parseOctaveUp(">");
                    break;
                case '<':
                    next(1);
                    parseOctaveDown("<");
                    break;
                default:
                    throw new ParseException(Lang.PARSER_FOUND_UNKNOWN_WORD, "" + c, fileno, lineno);
            }
        }
        return commands;
    }
    
    public void skipSpace() {
        while(!EOS()) {
            char c = getCur();
            if (c == ' ' || c == '\t' || c == '|') {
                next(1);
                continue;
            }
            break;
        }
    }
    
    private String getLengthParam() {
        String res = "";
        while (!EOS()) {
            skipSpace();
            char c = getChar();
            if (isNumberic(c) || c == '^' || c == '%' || c == '.') {
                res += c;
                next(1);
                continue;
            }
            back(1);
            break;
        }
        return res;
    }
    
    public String getStringRange() {
        return getRange('{', '}');
    }
    public String getParethesisRange() {
        return getRange('(', ')');
    }
    
    private String getRange(char ch_begin, char ch_end) {
        String res = "";
        int nest = 0;
        skipSpace();
        // check begin
        if (ch_begin == getCur()) {
            nest++;
        }
        else {
            return "";
        }
        // get range
        while (!EOS()) {
            char c = getChar();
            if (c == ch_end) {
                nest--;
                if (nest <= 0) {
                    break;
                }
            }
            if (c == ch_begin) {
                nest++;
            }
            res += c;
        }
        return res;
    }
    
    private void parseNote(char c) {
        SCommandNote cmd = new SCommandNote();
        commands.addCommand(cmd);
        cmd.note = c;
        next(1); // skip NOTE
        // check sharp
        while (!EOS()) {
            char sf = getCur();
            if (sf == '+' || sf == '#') {
                cmd.note++;
                next(1);
                continue;
            }
            if (sf == '-') {
                cmd.note--;
                next(1);
                continue;
            }
            break;
        }
        // check length
        cmd.length = getLengthParam();
        skipSpace();
        if (!startsWith(',')) return;
        // check gate
        skipSpace();
        if (startsWith('%')) {
            cmd.gate_stepmode = true;
            next(1);
        }
        cmd.gate = getInt(SCommand.DEFAULT_PARAM_I);
        skipSpace();
        if (!startsWith(',')) return;
        // check veclocity
        skipSpace();
        char pm = getCur();
        if (pm == '+') { cmd.velocity_mode = SCommand.VALUE_RELATIVE; next(1); }
        else if (pm == '-') { cmd.velocity_mode = SCommand.VALUE_RELATIVE; } // DO NOT INCREMENT POINTER
        cmd.velocity = getInt(SCommand.DEFAULT_PARAM_I);
        skipSpace();
        if (!startsWith(',')) return;
        // check timing
        skipSpace();
        pm = getCur();
        if (pm == '+') { cmd.timing_mode = SCommand.VALUE_RELATIVE; next(1); }
        else if (pm == '-') { cmd.timing_mode = SCommand.VALUE_RELATIVE; } // DO NOT INCREMENT POINTER
        cmd.timing = getInt(SCommand.DEFAULT_PARAM_I);
    }
    
    private void parseRest() {
        next(1); // skip 'r'
        SCommandRest cmd = new SCommandRest();
        commands.addCommand(cmd);
        // check length
        cmd.length = getLengthParam();
    }
    
    private void parseProgramChange(String cname) 
        throws ParseException {
        SCommandProgramChange cmd = new SCommandProgramChange();
        commands.addCommand(cmd);
        // check
        cmd.program_no = getInt(-1) - 1;
        if (cmd.program_no < 0) {
            throw new ParseException(Lang.PARSER_COMMAND_INVALID_ARG, cname, fileno, lineno);
        }
    }
    
    private void checkFlag(SCommandValue cmd) {
        // (+|-)
        skipSpace();
        char c = getCur();
        if (c == '+' || c == '-') {
            cmd.value_mode = SCommand.VALUE_RELATIVE;
            next(1);
        }
    }
    private void checkValue(SCommandValue cmd, String cname)
        throws ParseException
    {
        // check
        cmd.value = getInt(Integer.MIN_VALUE);
        if (cmd.value == Integer.MIN_VALUE) {
            throw new ParseException(Lang.PARSER_COMMAND_INVALID_ARG, 
                    cname, fileno, lineno);
        }
    }
    
    private void parseOctave(String cname) 
        throws ParseException {
        SCommandOctave cmd = new SCommandOctave();
        commands.addCommand(cmd);
        checkFlag(cmd);
        checkValue(cmd, cname);
    }
    
    private void parseVelocity(String cname) 
        throws ParseException {
        SCommandVelocity cmd = new SCommandVelocity();
        commands.addCommand(cmd);
        checkFlag(cmd);
        checkValue(cmd, cname);
    }
    private void parseLength(String cname) 
        throws ParseException {
        SCommandLength cmd = new SCommandLength();
        commands.addCommand(cmd);
        checkFlag(cmd);
        // length
        cmd.len_str = getLengthParam();
        if (cmd.len_str == "") {
            throw new ParseException(Lang.PARSER_COMMAND_INVALID_ARG, 
                    cname, fileno, lineno);
        }
    }
    private void parseOctaveUp(String cname) {
        SCommandOctaveUp cmd = new SCommandOctaveUp();
        commands.addCommand(cmd);
    }
    private void parseOctaveDown(String cname) {
        SCommandOctaveDown cmd = new SCommandOctaveDown();
        commands.addCommand(cmd);
    }
}
