/*
 * Tokenizer.java
 *
 * Created on 2007/10/09, 0:43
 */

package com.kujirahand.utils;

public class Tokenizer {
    
    protected String source;
    protected int position;
    
    public Tokenizer(String source) {
        this.source = source;
        this.position = 0;
    }
    public void setSource(String source) {
        this.source = source;
        this.position = 0;
    }
    /**
     * End Of String ?
     */
    public boolean EOS() {
        return (position >= source.length());
    }
    
    public char getChar() {
        char res = source.charAt(position++);
        return res;
    }
    
    public char getCur() {
        char res = source.charAt(position);
        return res;
    }
    
    public void back(int count) {
        position -= count;
    }
    public void next(int count) {
        position += count;
    }
    
    public String getString(int len) {
        String res = source.substring(position, position + len);
        position += len;
        return res;
    }
    
    public String getToEnd() {
        String res = source.substring(position);
        position = source.length();
        return res;
    }
    /**
     * get token to splitter
     */
    public String getToken(String splitter) {
        String res;
        int i = source.indexOf(splitter, position);
        if (i < 0) {
            res = source.substring(position);
            position = splitter.length();
            return res;
        }
        res = source.substring(position, i);
        position = i + splitter.length();
        return res;
    }
    public boolean startsWith(String word) {
        return source.startsWith(word, position);
    }
    public boolean startsWith(char ch) {
        if (EOS()) return false;
        return (ch == source.charAt(position));
    }
    /**
     * get token to variable name in C program
     */
    public String getCVarName() {
        String res = "";
        char c;
        if (EOS()) return "";
        // first letter
        c = getChar();
        if (c == '_' || isAlpha(c) || c > 255) {
            res += c;
        }
        else {
            back(1);
            return "";
        }
        // second letter
        while (!EOS()) {
            c = getChar();
            if (c == '_' || isAlphaNum(c) || c > 255) {
                res += c;
            }
            else {
                back(1);
                break;
            }
        }
        return res;
    }
    
    public int getInt(int default_value) {
        // get numeric string
        int len = 0;
        int ret = 0;
        int flg = 1;
        if (startsWith('-')) { flg = -1; }
        while (!EOS()) {
            char c = getChar();
            if (!isNumberic(c)) {
                back(1);
                break;
            }
            len++;
            ret = (ret * 10) + (c - '0');
        }
        if (len == 0) return default_value;
        return ret * flg;
    }
    
    public float getNumber(float default_value) {
        String res = "";
        // [-]*
        int flg = 1;
        if (startsWith('-')) { flg = -1; }
        // [0-9]+
        while (!EOS()) {
            char c = getChar();
            if (!isNumberic(c)) {
                back(1);
                break;
            }
            res += c;
        }
        if (res == "") { return default_value; }
        // [.]
        if (!startsWith('.')) {
            return Integer.parseInt(res);
        }
        res += ".";
        position++;
        // [0-9]+
        while (!EOS()) {
            char c = getChar();
            if (!isNumberic(c)) {
                back(1);
                break;
            }
            res += c;
        }
        return Float.parseFloat(res) * flg;
    }
    
    public void skipSpace() {
        while (!EOS()) {
            char c = getCur();
            if (c == ' ' || c == '\t') {
                next(1);
                continue;
            }
            break;
        }
    }
    
    public static boolean isNumberic(char c) {
        return ('0' <= c && c <= '9');
    }
    public static boolean isAlpha(char c) {
        return ('a' <= c && c <= 'z' || 'A' <= c && c <= 'Z');
    }
    public static boolean isAlphaNum(char c) {
        return (isNumberic(c) || isAlpha(c));
    }
    public static boolean isUpperCase(char c) {
        return ('A' <= c && c <= 'Z');
    }
    public static boolean isLowerCase(char c) {
        return ('a' <= c && c <= 'z');
    }
}
